@@ -21,17 +21,16 @@
     converted via Munch.to/fromDict().
 """
 
-__version__ = '2.3.0'
-VERSION = tuple(map(int, __version__.split('.')))
+import pkg_resources
+
+from .python3_compat import iterkeys, iteritems, Mapping, u
+
+__version__ = pkg_resources.get_distribution('munch').version
+VERSION = tuple(map(int, __version__.split('.')[:3]))
 
 __all__ = ('Munch', 'munchify', 'DefaultMunch', 'DefaultFactoryMunch', 'unmunchify')
 
 
-from collections import defaultdict
-
-
-from .python3_compat import *   # pylint: disable=wildcard-import
-
 
 class Munch(dict):
     """ A dictionary that provides attribute-style access.
@@ -72,6 +71,8 @@ class Munch(dict):
 
         See unmunchify/Munch.toDict, munchify/Munch.fromDict for notes about conversion.
     """
+    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
+        self.update(*args, **kwargs)
 
     # only called if k not found in normal places
     def __getattr__(self, k):
@@ -198,7 +199,7 @@ class Munch(dict):
 
         See https://docs.python.org/3.6/library/pickle.html.
         """
-        return self.toDict()
+        return {k: v for k, v in self.items()}
 
     def __setstate__(self, state):
         """ Implement a serializable interface used for pickling.
@@ -225,13 +226,37 @@ class Munch(dict):
     def copy(self):
         return type(self).fromDict(self)
 
+    def update(self, *args, **kwargs):
+        """
+        Override built-in method to call custom __setitem__ method that may
+        be defined in subclasses.
+        """
+        for k, v in iteritems(dict(*args, **kwargs)):
+            self[k] = v
+
+    def get(self, k, d=None):
+        """
+        D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None.
+        """
+        if k not in self:
+            return d
+        return self[k]
+
+    def setdefault(self, k, d=None):
+        """
+        D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D
+        """
+        if k not in self:
+            self[k] = d
+        return self[k]
+
 
 class AutoMunch(Munch):
     def __setattr__(self, k, v):
         """ Works the same as Munch.__setattr__ but if you supply
             a dictionary as value it will convert it to another Munch.
         """
-        if isinstance(v, dict) and not isinstance(v, (AutoMunch, Munch)):
+        if isinstance(v, Mapping) and not isinstance(v, (AutoMunch, Munch)):
             v = munchify(v, AutoMunch)
         super(AutoMunch, self).__setattr__(k, v)
 
@@ -266,7 +291,7 @@ class DefaultMunch(Munch):
         if k == '__default__':
             object.__setattr__(self, k, v)
         else:
-            return super(DefaultMunch, self).__setattr__(k, v)
+            super(DefaultMunch, self).__setattr__(k, v)
 
     def __getitem__(self, k):
         """ Gets key if it exists, otherwise returns the default value."""
@@ -280,7 +305,7 @@ class DefaultMunch(Munch):
 
         See https://docs.python.org/3.6/library/pickle.html.
         """
-        return (self.__default__, self.toDict())
+        return (self.__default__, {k: v for k, v in self.items()})
 
     def __setstate__(self, state):
         """ Implement a serializable interface used for pickling.
@@ -305,7 +330,7 @@ class DefaultMunch(Munch):
             type(self).__name__, self.__undefined__, dict.__repr__(self))
 
 
-class DefaultFactoryMunch(defaultdict, Munch):
+class DefaultFactoryMunch(Munch):
     """ A Munch that calls a user-specified function to generate values for
         missing keys like collections.defaultdict.
 
@@ -320,8 +345,8 @@ class DefaultFactoryMunch(defaultdict, Munch):
     """
 
     def __init__(self, default_factory, *args, **kwargs):
-        # pylint: disable=useless-super-delegation
-        super(DefaultFactoryMunch, self).__init__(default_factory, *args, **kwargs)
+        super(DefaultFactoryMunch, self).__init__(*args, **kwargs)
+        self.default_factory = default_factory
 
     @classmethod
     def fromDict(cls, d, default_factory):
@@ -336,6 +361,16 @@ class DefaultFactoryMunch(defaultdict, Munch):
         return '{0}({1}, {2})'.format(
             type(self).__name__, factory, dict.__repr__(self))
 
+    def __setattr__(self, k, v):
+        if k == 'default_factory':
+            object.__setattr__(self, k, v)
+        else:
+            super(DefaultFactoryMunch, self).__setattr__(k, v)
+
+    def __missing__(self, k):
+        self[k] = self.default_factory()
+        return self[k]
+
 
 # While we could convert abstract types like Mapping or Iterable, I think
 # munchify is more likely to "do what you mean" if it is conservative about
@@ -363,12 +398,48 @@ def munchify(x, factory=Munch):
 
         nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
     """
-    if isinstance(x, dict):
-        return factory((k, munchify(v, factory)) for k, v in iteritems(x))
-    elif isinstance(x, (list, tuple)):
-        return type(x)(munchify(v, factory) for v in x)
-    else:
-        return x
+    # Munchify x, using `seen` to track object cycles
+    seen = dict()
+
+    def munchify_cycles(obj):
+        # If we've already begun munchifying obj, just return the already-created munchified obj
+        try:
+            return seen[id(obj)]
+        except KeyError:
+            pass
+
+        # Otherwise, first partly munchify obj (but without descending into any lists or dicts) and save that
+        seen[id(obj)] = partial = pre_munchify(obj)
+        # Then finish munchifying lists and dicts inside obj (reusing munchified obj if cycles are encountered)
+        return post_munchify(partial, obj)
+
+    def pre_munchify(obj):
+        # Here we return a skeleton of munchified obj, which is enough to save for later (in case
+        # we need to break cycles) but it needs to filled out in post_munchify
+        if isinstance(obj, Mapping):
+            return factory({})
+        elif isinstance(obj, list):
+            return type(obj)()
+        elif isinstance(obj, tuple):
+            type_factory = getattr(obj, "_make", type(obj))
+            return type_factory(munchify_cycles(item) for item in obj)
+        else:
+            return obj
+
+    def post_munchify(partial, obj):
+        # Here we finish munchifying the parts of obj that were deferred by pre_munchify because they
+        # might be involved in a cycle
+        if isinstance(obj, Mapping):
+            partial.update((k, munchify_cycles(obj[k])) for k in iterkeys(obj))
+        elif isinstance(obj, list):
+            partial.extend(munchify_cycles(item) for item in obj)
+        elif isinstance(obj, tuple):
+            for (item_partial, item) in zip(partial, obj):
+                post_munchify(item_partial, item)
+
+        return partial
+
+    return munchify_cycles(x)
 
 
 def unmunchify(x):
@@ -388,12 +459,49 @@ def unmunchify(x):
 
         nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
     """
-    if isinstance(x, dict):
-        return dict((k, unmunchify(v)) for k, v in iteritems(x))
-    elif isinstance(x, (list, tuple)):
-        return type(x)(unmunchify(v) for v in x)
-    else:
-        return x
+
+    # Munchify x, using `seen` to track object cycles
+    seen = dict()
+
+    def unmunchify_cycles(obj):
+        # If we've already begun unmunchifying obj, just return the already-created unmunchified obj
+        try:
+            return seen[id(obj)]
+        except KeyError:
+            pass
+
+        # Otherwise, first partly unmunchify obj (but without descending into any lists or dicts) and save that
+        seen[id(obj)] = partial = pre_unmunchify(obj)
+        # Then finish unmunchifying lists and dicts inside obj (reusing unmunchified obj if cycles are encountered)
+        return post_unmunchify(partial, obj)
+
+    def pre_unmunchify(obj):
+        # Here we return a skeleton of unmunchified obj, which is enough to save for later (in case
+        # we need to break cycles) but it needs to filled out in post_unmunchify
+        if isinstance(obj, Mapping):
+            return dict()
+        elif isinstance(obj, list):
+            return type(obj)()
+        elif isinstance(obj, tuple):
+            type_factory = getattr(obj, "_make", type(obj))
+            return type_factory(unmunchify_cycles(item) for item in obj)
+        else:
+            return obj
+
+    def post_unmunchify(partial, obj):
+        # Here we finish unmunchifying the parts of obj that were deferred by pre_unmunchify because they
+        # might be involved in a cycle
+        if isinstance(obj, Mapping):
+            partial.update((k, unmunchify_cycles(obj[k])) for k in iterkeys(obj))
+        elif isinstance(obj, list):
+            partial.extend(unmunchify_cycles(v) for v in obj)
+        elif isinstance(obj, tuple):
+            for (value_partial, value) in zip(partial, obj):
+                post_unmunchify(value_partial, value)
+
+        return partial
+
+    return unmunchify_cycles(x)
 
 
 # Serialization
@@ -413,7 +521,14 @@ try:
         """
         return json.dumps(self, **options)
 
+    def fromJSON(cls, stream, *args, **kwargs):
+        """ Deserializes JSON to Munch or any of its subclasses.
+        """
+        factory = lambda d: cls(*(args + (d,)), **kwargs)
+        return munchify(json.loads(stream), factory=factory)
+
     Munch.toJSON = toJSON
+    Munch.fromJSON = classmethod(fromJSON)
 
 except ImportError:
     pass
@@ -468,8 +583,13 @@ try:
         """
         return dumper.represent_mapping(u('!munch.Munch'), data)
 
-    yaml.add_constructor(u('!munch'), from_yaml)
-    yaml.add_constructor(u('!munch.Munch'), from_yaml)
+    for loader_name in ("BaseLoader", "FullLoader", "SafeLoader", "Loader", "UnsafeLoader", "DangerLoader"):
+        LoaderCls = getattr(yaml, loader_name, None)
+        if LoaderCls is None:
+            # This code supports both PyYAML 4.x and 5.x versions
+            continue
+        yaml.add_constructor(u('!munch'), from_yaml, Loader=LoaderCls)
+        yaml.add_constructor(u('!munch.Munch'), from_yaml, Loader=LoaderCls)
 
     SafeRepresenter.add_representer(Munch, to_yaml_safe)
     SafeRepresenter.add_multi_representer(Munch, to_yaml_safe)
@@ -501,11 +621,13 @@ try:
         else:
             return yaml.dump(self, **opts)
 
-    def fromYAML(*args, **kwargs):
-        return munchify(yaml.load(*args, **kwargs))
+    def fromYAML(cls, stream, *args, **kwargs):
+        factory = lambda d: cls(*(args + (d,)), **kwargs)
+        loader_class = kwargs.pop('Loader', yaml.FullLoader)
+        return munchify(yaml.load(stream, Loader=loader_class), factory=factory)
 
     Munch.toYAML = toYAML
-    Munch.fromYAML = staticmethod(fromYAML)
+    Munch.fromYAML = classmethod(fromYAML)
 
 except ImportError:
     pass
